""" Common LDAP/LDIF helpers """

import logging
import re


log = logging.getLogger(__name__)
__version__ = '0.1.0'
_get_password_hash_type = re.compile('^{([^}]+)}.+')


def decode_value(options, value):
    """ Decode an attribute value from LDIF file """
    if isinstance(value, bytes):
        return value.decode(options.encoding, options.decoding_policy)
    return value


def get_entry_attr(options, entry, attr, all_values=True, default=None, decode=True):  # pylint: disable=too-many-arguments
    """ Get attribute value(s) from an entry of LDIF file """
    matched_attrs = [a for a in entry if a.lower() == attr.lower()]
    values = entry.get(matched_attrs[0]) if len(matched_attrs) == 1 else None
    if not isinstance(values, list) or not values:
        return default if not all_values or default is not None else []
    if decode:
        values = [decode_value(options, value) for value in values]
    return values if all_values else values[0]


def get_password_hash_type(pwd):
    """ Get password hash type """
    matched = _get_password_hash_type.match(pwd)
    if not matched:
        log.debug("get_password_hash_type(%s): not match => CLEAR", pwd)
        return 'CLEAR'
    hash_type = matched.group(1).upper()
    log.debug("get_password_hash_type(%s): %s", pwd, hash_type)
    if hash_type == 'CRYPT':
        if pwd[0:10].endswith('$1$'):
            hash_type = 'CRYPT-MD5'
        elif pwd[7:10] == '$2$' or pwd[7:11] in ('$2a$', '$2b$', '$2x$', '$2y$'):
            hash_type = 'CRYPT-BCRYPT'
        elif pwd[7:10] == '$3$':
            hash_type = 'CRYPT-NTHASH'
        elif pwd[7:10] == '$5$':
            hash_type = 'CRYPT-SSHA256'
        elif pwd[7:10] == '$6$':
            hash_type = 'CRYPT-SSHA512'
        elif pwd[7:12] == '$md5,':
            hash_type = 'CRYPT-SOLARIS-MD5'
        elif pwd[7:13] == '$sha1$':
            hash_type = 'CRYPT-PBKDF1-SHA1'
        elif pwd[7:8] == '_':
            hash_type = 'CRYPT-BSD-DES'
        else:
            hash_type = 'CRYPT-DES'
        log.debug("get_password_hash_type(%s): CRYPT sub type : %s", pwd, hash_type)

    return hash_type


def is_clear_password(value):
    """ Check if specified password is clear """
    hash_type = _get_password_hash_type.match(value)
    if not hash_type:
        return True
    return False
