""" Filtering stuff """

import argparse
import logging
import re

from ldap.dn import explode_dn

from mylib.pbar import Pbar

from ldap_audit_tools import get_entry_attr

log = logging.getLogger(__name__)
scopes = ('base', 'one', 'sub')


def scope_option_type(value):
    """ LDIF file scope option type as handle by argparse """
    if value not in scopes:
        try:
            if ':' in value:
                parts = value.split(':')
                if len(parts) > 2:
                    raise ValueError()
                value = (
                    int(parts[0]) if parts[0] else False,
                    int(parts[1]) if parts[1] else False
                )
            else:
                value = int(value)
                if value < 0:
                    raise ValueError('Numeric scope must be a positive integer')
        except ValueError as exc:
            raise argparse.ArgumentTypeError(
                'Invalid scope, must be an positive integer, a min/max levels '
                f'(format: min:max) or one of following: {",".join(scopes)}'
            ) from exc
    return value


def audit_filter_option_type(raw_value):
    """ Audit filter option type as handle by argparse """
    try:
        value = AuditFilter(raw_value)
    except InvalidAuditFilter as exc:
        raise argparse.ArgumentTypeError(f"Invalid filter '{raw_value}': {exc}") from exc
    return value


def add_filtering_options(parser):
    """ Add filtering options to an argparse.ArgumentParser instance """

    filter_opts = parser.add_argument_group("Filtering options")

    filter_opts.add_argument(
        '-b', '--base',
        action="store",
        dest="base",
        help="Base DN of search",
        type=str,
        default=None
    )

    filter_opts.add_argument(
        '-B', '--exclude-base',
        action="store_true",
        dest="exclude_base",
        help="Exclude base DN object"
    )

    filter_opts.add_argument(
        '-s', '--scope',
        action="store",
        dest="scope",
        help=(
            'Scope of the search (must be a postive integer corresponding to max search depth or ' f'one of the following: {",".join(scopes)})'),
        type=scope_option_type,
    )

    filter_opts.add_argument(
        '-f', '--filters',
        action="append",
        type=audit_filter_option_type,
        dest="filters",
        help=(
            "Filters on object attribute values. Format: [attribute name][!][operator][pattern]. "
            "Supported operators: = / > / < / >= / <= / ~= (regex pattern) / == (equality with "
            "another attribute). You could prefix the operator with an exclamation point to  "
            "inverse the logical and you could filter on count attribute values using count(attr) "
            "as attribute name. Multiple filter could be specified by repeating this parameter."
        )
    )

    filter_opts.add_argument(
        '-O', '--or',
        action="store_true",
        dest='combine_filters_with_or',
        help="Combine filters with an OR logical operator instead of AND one."
    )

    return filter_opts


def check_filtering_options(parser, options):
    """ Check validty of filtering options in an argparse.Namespace instance """
    if options.scope is not None and not options.base:
        parser.error("To use -s/--scope parameter, you must specify a base DN using -b/--base")

    if options.exclude_base and not options.base:
        parser.error("To use -B/--exclude-base parameter, you must specify a base DN using -b/--base")

    if options.exclude_base and options.scope in ['base', 0]:
        parser.error("Exclude base filter and base (or zero) scope are antinomic")


class InvalidAuditFilter(Exception):
    """ Exception raised on invalid audit filter value detected """


class AuditFilter:
    """ Class to implement a audit filter rule """

    __parse_audit_filter = re.compile(
        r'^((?P<attr_name>[a-zA-Z0-9\;]+)|count\((?P<count_attr_name>[a-zA-Z0-9\;]+)\))(?P<reverse>\!?)(?P<op>==|=|~=|>=|<=|>|<)(?P<pattern>.*)$')

    def __init__(self, raw_value):
        # Try to parse the audit filter
        matched = self.__parse_audit_filter.match(raw_value)
        if not matched:
            raise InvalidAuditFilter(f'Invalid filter "{raw_value}"')

        pattern = matched.group('pattern')

        # Handle specific operation need for some operators
        if matched.group('op') == '~=':
            # Regex filter: try to compile provided pattern
            try:
                pattern = re.compile(pattern)
            except re.error as exc:
                raise InvalidAuditFilter(
                    f'Invalid filter "{raw_value}": fail to compile regex ({exc})'
                ) from exc

        elif matched.group('op') in ('<', '<=', '>', '>='):
            # Numeric check: try to cast provided pattern value
            try:
                pattern = float(pattern)
            except ValueError as exc:
                raise InvalidAuditFilter(
                    f'Invalid filter "{raw_value}": {exc}'
                ) from exc

        self.raw_value = raw_value
        self.attr = (matched.group('attr_name') or matched.group('count_attr_name')).lower()
        self.count = bool(matched.group('count_attr_name'))
        self.reverse = bool(matched.group('reverse'))
        self.operator = matched.group('op')
        self.pattern = pattern

    def __match_count(self, value):
        """
        Check if value match: count check

        :param value: The value to check

        return True if entry matched, False otherwise
        """
        # Try to cast value as float
        try:
            value = float(value)
        except ValueError:
            return False

        if self.operator == '=':
            return value == self.pattern

        if self.operator == '>':
            return value > self.pattern

        if self.operator == '>=':
            return value >= self.pattern

        if self.operator == '<':
            return value < self.pattern

        if self.operator == '<=':
            return value <= self.pattern

        raise ValueError(f'Invalid operator "{self.operator}"')

    def __match_equal(self, value):
        """
        Check if value match: equal check

        :param value: The value to check

        return True if entry matched, False otherwise
        """
        if self.pattern == '*':
            return True

        if self.pattern.startswith('*') and self.pattern.endswith('*'):
            return self.pattern[1:-1] in value

        if self.pattern.startswith('*'):
            return value.endswith(self.pattern[1:])

        if self.pattern.endswith('*'):
            return value.startswith(self.pattern[0:-1])

        return self.pattern == value

    def __match_regex(self, value):
        """
        Check if value match: regex check

        :param value: The value to check

        return True if entry matched, False otherwise
        """
        return self.pattern.match(value)

    def __match_all_attr_values(self, options, entry_dn, entry, match_method):
        """
        Check if at least one attribute value match with the provided match method

        :param options: The options of the current script (argparse.Namespace)
        :param entry_dn: The entry DN
        :param entry: The entry object
        :param match_method: The match method

        return True if at least one attribute value matched, False otherwise
        """
        if self.attr.lower() == 'dn':
            return match_method(entry_dn)

        for value in get_entry_attr(options, entry, self.attr):
            if match_method(value):
                return True
        return False

    def match(self, options, entry_dn, entry):
        """
        Check if an LDIF entry match with this filter

        :param options: The options of the current script (argparse.Namespace)
        :param entry_dn: The entry DN
        :param entry: The entry's attributes as dict

        return True if entry matched, False otherwise
        """
        matched = False
        if isinstance(self.pattern, re.Pattern):
            matched = self.__match_all_attr_values(options, entry_dn, entry, self.__match_regex)

        elif self.count:
            matched = self.__match_count(
                len(get_entry_attr(options, entry, self.attr)))

        elif self.operator in ('<', '<=', '>', '>='):
            matched = self.__match_all_attr_values(options, entry_dn, entry, self.__match_count)

        elif self.operator == '=':
            matched = self.__match_all_attr_values(options, entry_dn, entry, self.__match_equal)

        elif self.operator == '==':
            matched = bool(
                set(
                    [entry_dn] if self.attr.lower() == 'dn' else
                    get_entry_attr(options, entry, self.attr)
                )
                & set(
                    [entry_dn] if self.pattern.lower() == 'dn' else
                    get_entry_attr(options, entry, self.pattern)
                )
            )

        else:
            raise Exception(f'Fail to handle filter "{self}"')

        if self.reverse:
            return not matched
        return matched

    def __repr__(self):
        """ Get string representation of an AuditFilter object """
        words = [f'count({self.attr})' if self.count else self.attr]
        if self.reverse:
            words.append('!')
        words.extend([self.operator, str(self.pattern)])
        return f'AuditFilter({" ".join(words)})'


# pylint: disable=too-many-return-statements
def match_entry_scope(entry_dn, base, scope, exclude_base=False):
    """
    Check if entry is matched against provided rules

    :param entry_dn: The DN of the object
    :param base: Base DN
    :param scope: The scope
    :param exclude_base: If True, the base DN object will not match

    return True if object matched, False otherwise
    """
    if base and not entry_dn.endswith(base):
        log.debug('Object %s not in base DN %s', entry_dn, base)
        return False

    if exclude_base and entry_dn == base:
        log.debug('Object %s excluded by -B/--exclude-base parameter', entry_dn)
        return False

    if scope == 'sub' or not scope:
        # Scope sub : always match
        return True

    if scope == 'base':
        if entry_dn == base:
            return True
        log.debug('Object %s not matched with scope base on %s', entry_dn, base)
        return False

    assert scope == 'one' or isinstance(scope, (int, tuple)), (
        f'Unsupported scope "{scope}" ({type(scope)})')

    if scope in ('one', 1):
        min_level = 0
        max_level = 1
        scope_label = 'one'
    elif isinstance(scope, int):
        min_level = 0
        max_level = scope
        scope_label = f'max={max_level}'
    else:
        min_level = scope[0]
        max_level = scope[1]
        if min_level and max_level:
            scope_label = f'min={min_level}/max={max_level}'
        elif min_level:
            scope_label = f'min={min_level}'
        else:
            scope_label = f'max={max_level}'
    level = len(explode_dn(entry_dn)) - len(explode_dn(base))
    if (
        (min_level is not False and level < min_level)
        or (max_level is not False and level > max_level)
    ):
        log.debug(
            'Object %s not matched with scope %s on %s',
            entry_dn, scope_label, base)
        return False
    return True


# pylint: disable=too-many-arguments
def match_entry(options, entry_dn, entry, base=None, exclude_base=False, scope=None,
                filters=None, combine_filters_with_or=False):
    """
    Check if entry is matched against provided rules

    :param entry_dn: The DN of the object
    :param entry: The LDIF entry of the object
    :param base: Base DN
    :param scope: The scope
    :param exclude_base: If True, the base DN object will not match
    :param filters: List of AuditFilter for filtering

    return True if object matched, False otherwise
    """
    if not match_entry_scope(entry_dn, base, scope, exclude_base=exclude_base):
        return False

    if filters:
        matched_filters = False
        for audit_filter in filters:
            if audit_filter.match(options, entry_dn, entry):
                matched_filters = True
            else:
                log.debug("%s: does not match with filter %s", entry_dn, audit_filter)
                if not combine_filters_with_or:
                    return False
        if combine_filters_with_or and not matched_filters:
            log.debug("%s: match with no filter", entry_dn)
            return False

    log.debug("%s : matched", entry_dn)
    return True


def get_matching_entries(options, entries, **kwargs):
    """ Retrieve matching entries """
    return [
        (entry_dn, entry)
        for entry_dn, entry in (entries.items() if isinstance(entries, dict) else entries)
        if match_entry(options, entry_dn, entry, **kwargs)
    ]


def get_matching_options(options):
    """
    Return matching options from argparse.Namespace

    :param options: The options of the current script (argparse.Namespace)

    return Dict of matching options
    """
    return dict(
        base=options.base, exclude_base=options.exclude_base,
        scope=options.scope, filters=options.filters,
        combine_filters_with_or=options.combine_filters_with_or
    )


def iter_on_entries(options, entries, handle_entry, count=None, context=None):
    """
    Iter on entries

    :param options: The options of the current script (argparse.Namespace)
    :param entries: Iterable entries
    :param count: Entries count (optional)
    :param handle_entry: The callback method to handle one entry
    :param context: Optional context object pass to each callback entry

    return void
    """
    if not context:
        context = dict(total_objects=0, matching_objects=0)
    else:
        assert isinstance(context, dict)

    context['total_objects'] = count if count is not None else len(entries)
    pbar = Pbar('Itering on entries', context['total_objects'], enabled=options.progress)

    context['matching_objects'] = 0
    print(get_matching_options(options))
    for entry_dn, entry in get_matching_entries(
            options, entries,
            **get_matching_options(options)):
        context['matching_objects'] += 1
        handle_entry(entry_dn, entry, context, options)
        pbar.increment()

    pbar.finish()
