""" CSV output file stuff """

import csv

from mylib.pbar import Pbar


def add_csv_export_options(parser):
    """ Add CSV export options to an argparse.ArgumentParser instance """

    csv_opts = parser.add_argument_group("CSV output options")

    csv_opts.add_argument(
        '--csv-delimiter',
        action='store',
        dest='csv_delimiter',
        help="CSV delimiter (default: ';')",
        default=';'
    )

    csv_opts.add_argument(
        '--csv-quotechar',
        action='store',
        dest='csv_quotechar',
        help="CSV quotechar (default: '\"')",
        default='"'
    )

    return csv_opts


#
# CSV Export helper
#
class CSVExport:
    """ Class to implement CSV exporting """

    def __init__(self, options, file_descriptor, **kwds):
        # Redirect output to a queue
        self.writer = csv.writer(
            file_descriptor,
            delimiter=options.csv_delimiter,
            quotechar=options.csv_quotechar,
            quoting=csv.QUOTE_MINIMAL,
            escapechar='\\',
            **kwds
        )

    def writerow(self, row):
        """ Write one row in CSV file """
        self.writer.writerow(row)

    def writerows(self, rows):
        """ Write multiple rows in CSV file """
        for row in rows:
            self.writerow(row)


def generate_csv_entries_output(options, entries, attributes_names):
    """ Generate an LDIF output file """
    attributes_names = sorted(attributes_names)
    generate_csv_output(
        options,
        headers=['dn'] + attributes_names,
        rows=[
            (
                [entry_dn] + [
                    '|'.join(attrs[attr])
                    if attr in attrs else ""
                    for attr in attributes_names
                ]
            )
            for entry_dn, attrs in entries.items()
        ]
    )


def generate_csv_output(options, headers, rows, msg=None):
    """ Generate an LDIF output file """
    with open(options.output, 'w', encoding=options.output_encoding) as file_descriptor:
        pbar = Pbar(
            msg if msg else f'Export result to "{options.output}"',
            len(rows),
            enabled=options.progress)
        output = CSVExport(options, file_descriptor)
        output.writerow(headers)
        for row in rows:
            output.writerow(row)
            pbar.increment()
        pbar.finish()
