""" Tool to detect clear password in LDAP file """

import logging

import argcomplete

from mylib.scripts.helpers import get_opts_parser
from mylib.scripts.helpers import init_logging

from ldap_audit_tools import get_entry_attr
from ldap_audit_tools import is_clear_password
from ldap_audit_tools.output import add_export_options
from ldap_audit_tools.output.csv import generate_csv_output
from ldap_audit_tools.filtering import add_filtering_options
from ldap_audit_tools.filtering import check_filtering_options
from ldap_audit_tools.filtering import iter_on_entries
from ldap_audit_tools.source import add_source_options
from ldap_audit_tools.source import check_source_options
from ldap_audit_tools.source import get_entries


log = logging.getLogger(__name__)
DEFAULT_ATTR = 'userPassword'


def handle_entry(entry_dn, entry, context, options):  # pylint: disable=unused-argument
    """ Handle one entry """
    detected = False
    for password in get_entry_attr(options, entry, options.attr):
        context['total_passwords'] += 1
        if is_clear_password(password):
            detected = True
            context['clear_passwords'] += 1

    if detected:
        context['obj_with_clear_passwords'].append(entry_dn)


def main(argv=None):
    """ Entry point of the script """
    parser = get_opts_parser(desc=__doc__, progress=True)

    add_source_options(parser)
    add_filtering_options(parser)
    add_export_options(parser, only_csv=True)

    parser.add_argument(
        action="store",
        dest="attr",
        help=f'Attribute name (default: {DEFAULT_ATTR})',
        type=str,
        nargs='?',
        default=DEFAULT_ATTR
    )

    argcomplete.autocomplete(parser)
    options = parser.parse_args(argv)
    check_source_options(parser, options)
    check_filtering_options(parser, options)

    init_logging(options, "Detect clear passwords from LDAP")

    context = dict(
        total_passwords=0,
        obj_with_clear_passwords=[],
        clear_passwords=0
    )
    entries = get_entries(options)
    iter_on_entries(options, entries, handle_entry, context=context)

    if not context['clear_passwords']:
        log.info(
            "No clear password detected on %d password(s) / %d object(s) checked",
            context['total_passwords'], context['matching_objects']
        )
        return True

    log.warning(
        '%d clear password(s) detected in %d object(s) (on %d password(s) / %d object(s) checked):\n - %s',
        context['clear_passwords'],
        len(context['obj_with_clear_passwords']),
        context['total_passwords'], context['matching_objects'],
        '\n - '.join(context['obj_with_clear_passwords'])
    )

    if options.output:
        generate_csv_output(
            options,
            headers=['DN'],
            rows=[
                [entry_dn]
                for entry_dn in context['obj_with_clear_passwords']
            ],
            msg=f'Export list of object(s) with clear passwords to "{options.output}"'
        )

    return False
