""" Tool to search entries from LDAP server of from LDIF file """

import logging

import argcomplete

from mylib.scripts.helpers import get_opts_parser
from mylib.scripts.helpers import init_logging

from ldap_audit_tools import get_entry_attr
from ldap_audit_tools.filtering import add_filtering_options
from ldap_audit_tools.filtering import check_filtering_options
from ldap_audit_tools.filtering import iter_on_entries
from ldap_audit_tools.output import add_export_options
from ldap_audit_tools.output import generate_entries_output
from ldap_audit_tools.source import add_source_options
from ldap_audit_tools.source import check_source_options
from ldap_audit_tools.source import get_entries


log = logging.getLogger(__name__)


def handle_entry(entry_dn, entry, context, options):  # pylint: disable=unused-argument
    """ Handle one entry """
    context['objects'][entry_dn] = {}
    for attr in entry:
        attr = attr.lower()
        if context['attributes'] and attr not in context['attributes']:
            continue
        if attr not in context['attributes_names']:
            context['attributes_names'].append(attr)
        context['objects'][entry_dn][attr] = get_entry_attr(options, entry, attr)


def main(argv=None):
    """ Entry point of the script """
    parser = get_opts_parser(desc=__doc__, progress=True)
    add_source_options(parser)
    add_filtering_options(parser)
    add_export_options(parser)

    parser.add_argument(
        '-M', '--max-length',
        action="store",
        dest="maxlen",
        help="Attributes's values max length",
        type=int,
        default=30
    )

    parser.add_argument(
        action="store",
        dest="attrs",
        help="Attribute(s) name",
        nargs='*',
        type=str
    )

    argcomplete.autocomplete(parser)
    options = parser.parse_args(argv)
    check_source_options(parser, options)
    check_filtering_options(parser, options)

    init_logging(options, "Search matching entries in LDIF")

    context = dict(
        attributes=[],
        attributes_names=[],
        objects={},
    )
    if options.attrs:
        context['attributes'].extend([attr.strip().lower() for attr in options.attrs])
    entries = get_entries(options)
    iter_on_entries(options, entries, handle_entry, context=context)

    log.info(
        '# %s maching object(s) found in LDIF file %s"',
        context['total_objects'], options.input)
    log.info('# Filter(s): %s', options.filters)
    log.info('# Base DN: %s', options.base if options.base else 'not specified')
    log.info('# Attribute(s): %s', options.attrs if options.attrs else "all")

    if options.output:
        generate_entries_output(options, context['objects'], context['attributes_names'])
    else:
        for entry_dn, attrs in context['objects'].items():
            print(f'dn: {entry_dn}')
            for attr, values in attrs.items():
                try:
                    for value in values:
                        print(
                            "{attr}: {value}".format(  # pylint: disable=consider-using-f-string
                                attr=attr,
                                value=(
                                    value if len(value) <= (options.maxlen + 5)
                                    else f'{value[:options.maxlen]}[...]'
                                )
                            )
                        )
                except Exception:  # pylint: disable=broad-except
                    log.error(
                        "Fail to export attribute %s value(s) of object %s",
                        attr, entry_dn, exc_info=True)
                    print(f'{attr}: [*** error exporting attribute value ***]')
            print()

        print(
            f'{context["matching_objects"]} matching object(s) on {str(context["total_objects"])}')
