""" Tool to list attribute nomenclature values from LDIF file """

import logging

import argcomplete

from mylib.scripts.helpers import get_opts_parser
from mylib.scripts.helpers import init_logging

from ldap_audit_tools import get_entry_attr
from ldap_audit_tools.output import add_export_options
from ldap_audit_tools.output.csv import generate_csv_output
from ldap_audit_tools.filtering import add_filtering_options
from ldap_audit_tools.filtering import check_filtering_options
from ldap_audit_tools.filtering import iter_on_entries
from ldap_audit_tools.source import add_source_options
from ldap_audit_tools.source import check_source_options
from ldap_audit_tools.source import get_entries


log = logging.getLogger(__name__)


def handle_entry(entry_dn, entry, context, options):
    """ Handle one entry """
    attr_values = get_entry_attr(options, entry, options.attr)
    if not attr_values:
        return

    context['total_objects_present'] += 1

    for value in attr_values:
        if value not in context['values']:
            context['values'][value] = []
            context['values_nb'][value] = 0
        if len(context['values'][value]) < options.nbexample:
            context['values'][value].append(entry_dn)
        context['values_nb'][value] += 1


def main(argv=None):
    """ Entry point of the script """
    parser = get_opts_parser(desc=__doc__, progress=True)
    add_source_options(parser)
    add_filtering_options(parser)
    add_export_options(parser, only_csv=True)

    parser.add_argument(
        action="store",
        dest="attr",
        help="Attribute name",
        type=str,
        default=None
    )

    parser.add_argument(
        '-N', '--nb-example',
        action="store",
        dest="nbexample",
        help="Number of object DN to retreived for each attribute value",
        type=int,
        default=3
    )

    parser.add_argument(
        '-S', '--sort-by-occurences',
        action="store_true",
        dest="sort_by_occurences",
        help="Sort result by number of occurences of the values"
    )

    parser.add_argument(
        '-R', '--reverse',
        action="store_true",
        dest="reverse",
        help="Reverse result ordering"
    )

    argcomplete.autocomplete(parser)
    options = parser.parse_args(argv)
    check_source_options(parser, options)
    check_filtering_options(parser, options)

    if not options.attr:
        parser.error("You must provide attribute name using -a/--attribute parameter.")

    init_logging(options, "List attribute nomenclature from LDIF")

    context = dict(
        values={},
        values_nb={},
        total_objects_present=0,
    )
    entries = get_entries(options)
    iter_on_entries(options, entries, handle_entry, context=context)

    print(f'Attribute {options.attr} values ({len(context["values"])} found) :')
    if options.sort_by_occurences:
        ordered_values = sorted(
            context['values_nb'].keys(),
            key=lambda x: context['values_nb'][x],
            reverse=options.reverse)
    else:
        ordered_values = sorted(
            context['values'].keys(),
            reverse=options.reverse)

    for value in ordered_values:
        if context['values'][value]:
            print(f' - "{value}" ({context["values_nb"][value]} occurence(s)):')
            print('    - %s' % '\n    - '.join(context["values"][value]))
        else:
            print(f' - "{value}" ({context["values_nb"][value]} occurence(s))')
    print(
        f'Attribute {options.attr} found on {context["total_objects_present"]} object(s) on '
        f'{context["matching_objects"]}')

    if options.output:
        generate_csv_output(
            options,
            headers=['Value', 'Nb occurence', 'Objects using this value'],
            rows=[
                [
                    value,
                    str(context['values_nb'][value])
                ] + context['values'][value]
                for value in ordered_values
            ],
            msg=f'Export nomenclature values to "{options.output}"'
        )
