""" Tool to list attribute duplicated values from LDIF file """

import logging

import argcomplete

from mylib.scripts.helpers import get_opts_parser
from mylib.scripts.helpers import init_logging

from ldap_audit_tools import get_entry_attr
from ldap_audit_tools.output import add_export_options
from ldap_audit_tools.output.csv import generate_csv_output
from ldap_audit_tools.filtering import add_filtering_options
from ldap_audit_tools.filtering import check_filtering_options
from ldap_audit_tools.filtering import iter_on_entries
from ldap_audit_tools.source import add_source_options
from ldap_audit_tools.source import check_source_options
from ldap_audit_tools.source import get_entries


log = logging.getLogger(__name__)


def handle_entry(entry_dn, entry, context, options):  # pylint: disable=unused-argument
    """ Handle one entry """
    for attr in context['attributes']:
        attr_values = get_entry_attr(options, entry, attr)
        if not attr_values:
            return

        context['total_objects_present'] += 1

        for value in attr_values:
            if value not in context['values']:
                context['values'][value] = [(entry_dn, attr)]
            else:
                context['values'][value].append((entry_dn, attr))
                if value not in context['duplicated_values']:
                    context['duplicated_values'].append(value)


def main(argv=None):
    """ Entry point of the script """
    parser = get_opts_parser(desc=__doc__, progress=True)

    add_source_options(parser)
    add_filtering_options(parser)
    add_export_options(parser, only_csv=True)

    parser.add_argument(
        action="store",
        dest="attrs",
        help="Attribute name(s). You could specify multiple attribute names.",
        type=str,
        nargs="+",
        default=None
    )

    parser.add_argument(
        '-S', '--sort-by-occurences',
        action="store_true",
        dest="sort_by_occurences",
        help="Sort result by number of duplicated values"
    )

    parser.add_argument(
        '-R', '--reverse',
        action="store_true",
        dest="reverse",
        help="Reverse result ordering"
    )

    argcomplete.autocomplete(parser)
    options = parser.parse_args(argv)
    check_source_options(parser, options)
    check_filtering_options(parser, options)

    if not options.attrs:
        parser.error("You must provide attribute name using -a/--attribute parameter.")

    init_logging(options, "List duplicated attributes values from LDIF")

    context = dict(
        attributes=[attr.strip() for attr in options.attrs],
        values={},
        duplicated_values=[],
        total_objects_present=0,
    )
    entries = get_entries(options)
    iter_on_entries(options, entries, handle_entry, context=context)

    print(
        f'{len(context["duplicated_values"])} duplicated attribute '
        f'{",".join(context["attributes"])} values found :')

    if options.sort_by_occurences:
        ordered_values = sorted(
            context['duplicated_values'],
            key=lambda x: len(context['values'][x]),
            reverse=options.reverse)
    else:
        ordered_values = sorted(
            context['duplicated_values'],
            reverse=options.reverse)

    for value in ordered_values:
        print(f' - "{value}" ({len(context["values"][value])} occurence(s):')
        print(
            '\n    - %s' % '\n    - '.join([
                entry_dn if len(context['attributes']) == 1
                else f'{entry_dn} ({attr})'
                for entry_dn, attr in context['values'][value]
            ])
        )
    print(
        f'Attribute(s) {"/".join(options.attrs)} found on {context["total_objects_present"]} object(s) on '
        f'{context["matching_objects"]}')

    if options.output:
        generate_csv_output(
            options,
            headers=['Value', 'Nb occurence', 'Objects using this value'],
            rows=[
                [value, str(len(context['values'][value]))]
                + [
                    entry_dn if len(context['attributes']) == 1
                    else f'{entry_dn} ({attr})'
                    for entry_dn, attr in context['values'][value]
                ]
                for value in ordered_values
            ],
            msg=f'Export duplicated values to "{options.output}"'
        )
