""" Tool to analyse objectClasses and attributes used in LDIF file """

import logging

import argcomplete

from mylib.scripts.helpers import get_opts_parser
from mylib.scripts.helpers import init_logging

from ldap_audit_tools import get_entry_attr
from ldap_audit_tools.filtering import add_filtering_options
from ldap_audit_tools.filtering import check_filtering_options
from ldap_audit_tools.filtering import iter_on_entries
from ldap_audit_tools.source import add_source_options
from ldap_audit_tools.source import check_source_options
from ldap_audit_tools.source import get_entries


log = logging.getLogger(__name__)


def handle_entry(entry_dn, entry, context, options):  # pylint: disable=too-many-branches
    """ Hande one LDIF entry """

    # Handle RDN attribute
    rdn_attr = entry_dn.strip().split('=')[0]
    if rdn_attr not in context['rdn_attrs']:
        context['rdn_attrs'][rdn_attr] = []
    context['rdn_attrs'][rdn_attr].append(entry_dn)

    # Handle objectclasses
    for objectclass in get_entry_attr(options, entry, 'objectClass'):
        if objectclass not in context['objectclasses']:
            context['objectclasses'][objectclass] = []
            context['objectclasses_nb'][objectclass] = 0
        if len(context['objectclasses'][objectclass]) < options.nbexample:
            context['objectclasses'][objectclass].append(entry_dn)
        context['objectclasses_nb'][objectclass] += 1

    # Check objectclasses
    if context['check_objectclasses']:
        for objectclass in context['check_objectclasses']:
            if objectclass not in get_entry_attr(options, entry, 'objectClass'):
                log.warning('Object %s does not have objectClass %s', entry_dn, objectclass)

    # Check attributes
    if context['check_attrs']:
        for attr in context['check_attrs']:
            if not get_entry_attr(options, entry, attr):
                log.warning('Object %s does not have attribute %s', entry_dn, attr)

    # Handle attributes values
    for attr in entry:
        if attr.lower() == 'objectclass':
            continue
        if attr not in context['attrs']:
            context['attrs'][attr] = {}
            context['attrs_nb'][attr] = 0
        if len(context['attrs'][attr]) < options.nbexample:
            context['attrs'][attr][entry_dn] = get_entry_attr(options, entry, attr)
        context['attrs_nb'][attr] += 1
        if attr not in context['attrs_multiple'] and len(entry[attr]) > 1:
            context['attrs_multiple'].append(attr)


def main(argv=None):
    """ Entry point of the script """
    parser = get_opts_parser(desc=__doc__, progress=True)

    add_source_options(parser)
    add_filtering_options(parser)

    parser.add_argument(
        '--check-objectclass',
        action="store",
        dest="check_objectclass",
        help="List of objectclass that object must have",
        type=str,
        nargs="*",
        default=[]
    )

    parser.add_argument(
        '--check-attribute',
        action="store",
        dest="check_attrs",
        help="List of attributes that object must have",
        type=str,
        nargs="*",
        default=[]
    )

    parser.add_argument(
        '-N', '--nb-example',
        action="store",
        dest="nbexample",
        help="Number of example to retreived for each attribute",
        type=int,
        default=3
    )

    parser.add_argument(
        '-L', '--example-max-length',
        action="store",
        dest="example_max_len",
        help="Max length of example value retreived for each attribute",
        type=int,
        default=15
    )

    argcomplete.autocomplete(parser)
    options = parser.parse_args(argv)
    check_source_options(parser, options)
    check_filtering_options(parser, options)

    init_logging(options, "List objectClasses and attributes from LDIF")

    context = dict(
        objectclasses={},
        objectclasses_nb={},
        attrs={},
        attrs_nb={},
        attrs_multiple=[],
        rdn_attrs={},
        check_objectclasses=[objectclass.strip() for objectclass in options.check_objectclass],
        check_attrs=[attr.strip() for attr in options.check_attrs],
    )
    entries = get_entries(options)
    iter_on_entries(options, entries, handle_entry, context=context)

    print(f'{context["matching_objects"]} matched object(s) found on {context["total_objects"]}')
    if len(context['rdn_attrs']) == 1:
        print(f'RDN attribute: {next(iter(context["rdn_attrs"]))} (always)')
    else:
        print("RDN attributes:")
        for rdn_attr in context['rdn_attrs']:
            print(
                f' - {rdn_attr} ({len(context["rdn_attrs"][rdn_attr])}/'
                f'{context["matching_objects"]}):')
            for entry_dn in context['rdn_attrs'][rdn_attr][:10]:
                print(f'   - {entry_dn}')
    print("ObjectClasses :")
    for objectclass in sorted(context['objectclasses'].keys()):
        print(
            ' - {0} ({1}):\n    - {2}'.format(
                objectclass,
                "always present"
                if context['objectclasses_nb'][objectclass] == context['matching_objects']
                else f'present on {context["objectclasses_nb"][objectclass]}/{context["matching_objects"]}',
                "\n    - ".join(context['objectclasses'][objectclass])
            )
        )

    print("Attributes :")
    for attr in sorted(context['attrs'].keys()):
        print(
            " - %s (%s, %s) :" % (  # pylint: disable=consider-using-f-string
                attr,
                "always present"
                if context['attrs_nb'][attr] == context['matching_objects']
                else f'present on {context["attrs_nb"][attr]}/{context["matching_objects"]}',
                "unique value" if attr not in context['attrs_multiple'] else "multiple values"
            )
        )
        for entry_dn, values in context['attrs'][attr].items():
            example_values = []
            for value in values:
                if len(value) > options.example_max_len:
                    example_values.append(value[:options.example_max_len] + "[...]")
                else:
                    example_values.append(value)
            # pylint: disable=consider-using-f-string
            print('    - {0} : "{1}"'.format(
                entry_dn,
                '", "'.join(example_values)
            ))
