""" Tool to list password hash types present in LDIF file """

import logging

import argcomplete

from mylib.scripts.helpers import get_opts_parser
from mylib.scripts.helpers import init_logging

from ldap_audit_tools import get_entry_attr
from ldap_audit_tools import get_password_hash_type
from ldap_audit_tools.filtering import add_filtering_options
from ldap_audit_tools.filtering import check_filtering_options
from ldap_audit_tools.filtering import iter_on_entries
from ldap_audit_tools.source import add_source_options
from ldap_audit_tools.source import check_source_options
from ldap_audit_tools.source import get_entries


log = logging.getLogger(__name__)
DEFAULT_ATTRIBUTE = 'userPassword'


def handle_entry(entry_dn, entry, context, options):
    """ Handle one entry """
    pwds = get_entry_attr(options, entry, options.attribute)
    if not pwds:
        return
    context['objects_with_password'] += 1
    context['passwords'] += len(pwds)
    entry_hash_types = []
    for pwd in pwds:
        hash_type = get_password_hash_type(pwd)
        if hash_type not in context['hash_types']:
            context['hash_types'][hash_type] = []
        if hash_type not in entry_hash_types:
            log.debug('%s: %s', entry_dn, hash_type)
            entry_hash_types.append(hash_type)
            context['hash_types'][hash_type].append(entry_dn)


def main(argv=None):
    """ Entry point of the script """
    parser = get_opts_parser(desc=__doc__, progress=True)
    add_source_options(parser)
    add_filtering_options(parser)

    parser.add_argument(
        action="store",
        dest="attribute",
        help=f'Specified password attribute (default: {DEFAULT_ATTRIBUTE})',
        type=str,
        nargs='?',
        default=DEFAULT_ATTRIBUTE
    )

    parser.add_argument(
        '-N', '--nb-example',
        action="store",
        dest="nbexample",
        help="Number of example to retreived for each attribute",
        type=int,
        default=3
    )

    argcomplete.autocomplete(parser)
    options = parser.parse_args(argv)
    check_source_options(parser, options)
    check_filtering_options(parser, options)

    init_logging(options, "List password hash types from LDIF")

    context = dict(
        hash_types={},
        passwords=0,
        objects_with_password=0,
    )
    entries = get_entries(options)
    iter_on_entries(options, entries, handle_entry, context=context)

    print(
        f'{context["passwords"]} password(s) found on {context["objects_with_password"]} '
        f'object(s) (on a total of {context["matching_objects"]} object(s))'
    )
    print("Hash password types:")
    for hash_type in sorted(context['hash_types'].keys()):
        if options.nbexample:
            print(f' - {hash_type} ({len(context["hash_types"][hash_type])} accounts):')
            print('   - %s' % "\n   - ".join(
                context['hash_types'][hash_type][:options.nbexample]))
        else:
            print(f' - {hash_type} ({len(context["hash_types"][hash_type])} accounts)')
