""" Tool to audit OpenLDAP configuration """

import logging
import re

import argcomplete
import ldap.dn

from mylib.scripts.helpers import get_opts_parser
from mylib.scripts.helpers import init_logging

from ldap_audit_tools import get_entry_attr
from ldap_audit_tools.filtering import AuditFilter
from ldap_audit_tools.filtering import get_matching_entries
from ldap_audit_tools.source import add_source_options
from ldap_audit_tools.source import check_source_options
from ldap_audit_tools.source import get_entries


log = logging.getLogger(__name__)
DEFAULT_BASEDN = 'cn=config'
_parse_olc_numbered_value = re.compile(r'^\{(?P<idx>-?[0-9]+)\}(?P<value>.*)$')


def parse_olc_numbered_value(value):
    """ Parse numbered value and return split index and value """
    matched = _parse_olc_numbered_value.match(value)
    if matched:
        return (matched.group('idx'), matched.group('value'))
    return (None, value)


def get_olc_unnumbered_value(value):
    """ Return unnumbered value """
    return parse_olc_numbered_value(value)[1]


def extended_return(result_lines, lines, line_prefix=None):
    """
    Extended return

    :param result_lines: The result lines list
    :param lines: If defined, it must be a list and the resulting lines will be use to extend this
                  list
    :param line_prefix: If set, all resulting lines will be prefix with this string
    """
    result_lines = result_lines if isinstance(result_lines, list) else [result_lines]
    if line_prefix:
        result_lines = [line_prefix + result_line for result_line in result_lines]
    if isinstance(lines, list):
        lines.extend(result_lines)
    return result_lines


# pylint: disable=too-many-arguments,too-many-locals
def audit_attr(options, obj, attr, label, multiple=False, split=False,
               value_format='%s', format_value=None,
               not_set='not set', default=None, only_if_set=False,
               line_return=False, line_prefix=None, dash=False, lines=None):
    """
    Retrieve and audit object attribute value and return formated audit result message
    as list of lines.

    :param options: The script options as argparse.Namespace
    :param obj: The object
    :param attr: The attribute name
    :param label: The label of the audited info (corresponding to the attribute values)
    :param multiple: If True, consider the attribute as multi-valued (optional, default: False)
    :param split: If set, attribute's value(s) will be split with the provided value
                  (optional and multiple must also be True, default: False)
    :param value_format: Format use to display one attribut value (optional, default: '%s')
    :param format_value: If set all attributes values will be formated using this function
                        (optional, could be use in addition of value_format: in this case, value_format will be use to format resulting value of format_value)
    :param not_set: The label to display if attribut is not set (optional, default: 'not set')
    :param default: The default value/label to add as suffix to not set label (optional)
    :param only_if_set: If True, the audit result of this attribute will be displayed only if
                        attribute is set (optional, default: False)
    :param line_return: If True, attribute's value(s) will be displayed one by line instead
                        of be concatenated with comma in one line (optional, default: False)
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param dash: If True, label will be prefix with '- ' (optional, default: False)
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list

    :return: list of resultings lines of the attribute audit
    """
    label = f'- {label}' if dash else label
    values = get_entry_attr(options, obj, attr, all_values=True)
    if multiple and split:
        splited_values = []
        for value in values:
            splited_values.extend(
                value.split(split) if isinstance(split, str)
                else value.split()
            )
        values = splited_values
    if not values:
        if only_if_set:
            return []
        if default:
            not_set += f' (default: {default})'
        return extended_return(f'{label}: {not_set}', lines, line_prefix)

    if format_value:
        values = [format_value(value) for value in values]

    if not multiple:
        return extended_return(f'{label}: {value_format % values[0]}', lines, line_prefix)

    if line_return:
        result = [f'{label}:']
        result.extend([f' - {value_format % value}' for value in values])
        return extended_return(result, lines, line_prefix)
    return extended_return(
        f'{label}: {", ".join([value_format % value for value in values])}',
        lines, line_prefix)


def audit_attrs(label, *attrs_lines, empty_label='Not configured', only_not_empty=False,
                lines=None):
    """
    Format the audit result of a group of attributes

    :param label: The name/label of this group of attributes
    :param attrs_lines: All other positionnal parameters will be considered as the resulting
                        lines of the audit of one attribute
    :param empty_label: Label to display if all provided audit attributes result are empty
                        (optional, default: 'Not configured')
    :param only_not_empty: If True, the result will display only if provided audit attributes
                           result are not empty (optional, default: False)
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list

    :return: list of resultings lines of the group attributes audit
    """
    result_lines = []
    for attr_lines in attrs_lines:
        if len(attr_lines) > 1:
            result_lines.append(f'  - {attr_lines[0]}')
            result_lines.extend([f'    {line}' for line in attr_lines[1:]])
        else:
            result_lines.extend([f'  - {line}' for line in attr_lines])
    lines = lines if isinstance(lines, list) else []
    if result_lines:
        lines.append(f'{label}:')
        lines.extend(result_lines)
    elif not only_not_empty:
        lines.append(f'{label}: {empty_label}')
    return lines


def get_rdn_value(dn_value):
    """ Retrieve RDN value """
    return ldap.dn.str2dn(dn_value)[0][0][1]


def audit_databases(options, entries, line_prefix=None, lines=None):
    """
    Audit databases and print result to stdout

    :param options: The script options as argparse.Namespace
    :param entries: The sources LDAP entries
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []
    for entry_dn, entry in get_matching_entries(
            options, entries, filters=[AuditFilter('objectclass=olcDatabaseConfig')]):
        db_rdn = get_entry_attr(options, entry, 'olcDatabase', all_values=False)
        db_idx, db_type = parse_olc_numbered_value(db_rdn)

        result_lines = [
            '',
            '',
            f'Database #{db_idx}:',
            f'  - Type: {db_type.upper()}'
        ]
        audit_attr(
            options, entry, 'olcSuffix', 'Suffix', line_prefix='  - ', lines=result_lines)
        audit_attr(
            options, entry, 'olcDbDirectory', 'Storage directory',
            line_prefix='  - ', lines=result_lines)
        if db_type == 'mdb':
            audit_attr(
                options, entry, 'olcDbMaxSize', 'Maximum size (MDB)', default=10485760,
                value_format='%s bytes', line_prefix='  - ', lines=result_lines)

        audit_attr(
            options, entry, 'olcRootDN', 'Root DN', line_prefix='  - ', lines=result_lines)
        audit_attr(
            options, entry, 'olcRootPW', 'Root Password', line_prefix='  - ', lines=result_lines)

        def format_acl(acl):
            acl = re.sub(r'\s+by ', '\n         by ', acl)
            # Add line return after each attrs / filter clauses
            acl = re.sub(r'\s+(attrs|filter)=', r'\n         \1=', acl)
            # Add line return after each continue, break or stop clauses
            acl = re.sub(r'\s+(continue|break|stop)', r'\n         \1', acl)
            return acl
        audit_attr(
            options, entry, 'olcAccess', 'ACLs', line_return=True, lines=result_lines,
            not_set='no configured ACL', multiple=True, format_value=format_acl)

        def format_limit(limit):
            limit = re.sub(r'\s+by ', '\n         by ', limit)
            # Add line return after each size and time clauses
            limit = re.sub(r'\s+(size|time)\.', r'\n         \1.', limit)
            return limit
        audit_attr(
            options, entry, 'olcLimits', 'Limits', line_return=True, lines=result_lines,
            not_set='no configured limit', multiple=True, format_value=format_limit)

        indexes = {}
        for value in get_entry_attr(options, entry, 'olcDbIndex'):
            attrs, index_types = value.split()
            index_types = index_types.split(',')
            for attr in attrs.split(','):
                if attr not in indexes:
                    indexes[attr] = []
                indexes[attr].extend(index_types)
        if indexes:
            result_lines.append('Indexes:')
            result_lines.extend([
                f'  - {attr}: {", ".join(indexes[attr])}'
                for attr in sorted(indexes.keys())
            ])
        else:
            result_lines.append('Indexes: no configured index')

        overlays = get_matching_entries(
            options, entries, base=entry_dn, filters=[AuditFilter('objectclass=olcOverlayConfig')])
        if overlays:
            result_lines.append('Overlay(s):')
            for overlay_dn, overlay in overlays:
                overlay_rdn = get_rdn_value(overlay_dn)
                overlay_idx, overlay_type = parse_olc_numbered_value(overlay_rdn)
                result_lines.append(f'  - #{overlay_idx} - {overlay_type}')

                try:
                    globals()[f'audit_overlay_{overlay_type}'](
                        options, overlay_dn, overlay, line_prefix='    ', lines=result_lines)
                except KeyError:
                    pass
        else:
            result_lines.append('Overlay: no overlay configured')
        extended_return(result_lines, lines, line_prefix)


# pylint: disable=unused-argument
def audit_overlay_dyngroup(options, overlay_dn, overlay, line_prefix=None, lines=None):
    """
    Audit overlay dyngroup

    :param options: The script options as argparse.Namespace
    :param overlay_dn: The overlay DN
    :param overlay: The overlay entry
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []

    old_attr_pairs = get_entry_attr(options, overlay, 'olcDGAttrPair')
    attr_pairs = get_entry_attr(options, overlay, 'olcDynGroupAttrPair')
    if not old_attr_pairs and not attr_pairs:
        result_lines.append('- Attribute pair: not set')
    else:
        result_lines.append('- Attribute pair(s):')
        result_lines.extend([
            f'  - {attr_pair}'
            for attr_pair in (old_attr_pairs + attr_pairs)
        ])
        if old_attr_pairs:
            result_lines.append(
                '  => Caution: attribute olcDGAttrPair is deprecated since '
                'OpenLDAP 2.5 and replace by olcDynGroupAttrPair (and objectClass '
                'olcDynGroupConfig)')
    return extended_return(result_lines, lines, line_prefix)


# pylint: disable=unused-argument
def audit_overlay_dynlist(options, overlay_dn, overlay, lines=None, line_prefix=None):
    """
    Audit overlay dynlist

    :param options: The script options as argparse.Namespace
    :param overlay_dn: The overlay DN
    :param overlay: The overlay entry
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []
    audit_attr(
        options, overlay, 'olcDlAttrSet', 'Attribute set(s)', multiple=True,
        not_set='not set (/!\\ BUT REQUIRED /!\\)',
        dash=True, line_return=True, lines=result_lines)
    return extended_return(result_lines, lines, line_prefix)


# pylint: disable=unused-argument
def audit_overlay_constraint(options, overlay_dn, overlay, lines=None, line_prefix=None):
    """
    Audit overlay constraint

    :param options: The script options as argparse.Namespace
    :param overlay_dn: The overlay DN
    :param overlay: The overlay entry
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []
    audit_attr(
        options, overlay, 'olcConstraintAttribute', 'Constraint(s)', multiple=True,
        dash=True, line_return=True, lines=result_lines)
    return extended_return(result_lines, lines, line_prefix)


# pylint: disable=unused-argument
def audit_overlay_unique(options, overlay_dn, overlay, lines=None, line_prefix=None):
    """
    Audit overlay unique

    :param options: The script options as argparse.Namespace
    :param overlay_dn: The overlay DN
    :param overlay: The overlay entry
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []
    audit_attr(
        options, overlay, 'olcUniqueURI', 'Unique URI(s)', multiple=True,
        not_set='not set (/!\\ BUT REQUIRED /!\\)',
        dash=True, line_return=True, lines=result_lines)
    audit_attrs(
        'Legacy parameters that should be converted as unique URI',
        audit_attr(
            options, overlay, 'olcUniqueBase', 'Base DN', only_if_set=True,
            dash=True, line_prefix='  ', lines=result_lines),
        audit_attr(
            options, overlay, 'olcUniqueAttribute', 'Ignored attribute(s)', only_if_set=True,
            multiple=True, split=True,
            dash=True, line_prefix='  ', lines=result_lines),
        audit_attr(
            options, overlay, 'olcUniqueStrict', 'Strict', only_if_set=True,
            dash=True, line_prefix='  ', lines=result_lines),
        only_not_empty=True, lines=result_lines
    )
    return extended_return(result_lines, lines, line_prefix)


# pylint: disable=unused-argument
def audit_overlay_memberof(options, overlay_dn, overlay, lines=None, line_prefix=None):
    """
    Audit overlay memberof

    :param options: The script options as argparse.Namespace
    :param overlay_dn: The overlay DN
    :param overlay: The overlay entry
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []
    audit_attr(
        options, overlay, 'olcMemberOfGroupOC', 'Group objectClass', default='groupOfNames',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcMemberOfMemberAD', 'Group member attribute', default='member',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcMemberOfMemberOfAD', 'MemberOf attribute', default='memberOf',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcMemberOfRefInt', 'Refint enabled',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcMemberOfDN', 'Modifier DN', default='root',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcMemberOfDangling', 'Dangling policy', default='ignore',
        dash=True, lines=result_lines)
    # FIXME: detect and add notice on syncrepl context
    return extended_return(result_lines, lines, line_prefix)


# pylint: disable=unused-argument
def audit_overlay_ppolicy(options, overlay_dn, overlay, lines=None, line_prefix=None):
    """
    Audit overlay ppolicy

    :param options: The script options as argparse.Namespace
    :param overlay_dn: The overlay DN
    :param overlay: The overlay entry
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []
    audit_attr(
        options, overlay, 'olcPPolicyDefault', 'Default policy', not_set='no default policy',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcPPolicyHashCleartext', 'Hash clear text password', default='FALSE',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcPPolicyUseLockout', 'Use lockout', default='FALSE',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcPPolicyForwardUpdates', 'Forward updates', default='FALSE',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcPPolicySendNetscapeControls', 'Send netscape controls',
        default='FALSE', dash=True, lines=result_lines)
    return extended_return(result_lines, lines, line_prefix)


def audit_overlay_syncprov(options, overlay_dn, overlay, lines=None, line_prefix=None):
    """
    Audit overlay syncprov

    :param options: The script options as argparse.Namespace
    :param overlay_dn: The overlay DN
    :param overlay: The overlay entry
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []

    def format_checkpoint(checkpoint_value):
        checkpoint_ops, checkpoint_mins = checkpoint_value.split()
        return f'{checkpoint_ops} operation(s) / {checkpoint_mins} minute(s)'
    audit_attr(
        options, overlay, 'olcSpCheckpoint', 'Checkpoint', default='disabled',
        dash=True, format_value=format_checkpoint, lines=result_lines)
    audit_attr(
        options, overlay, 'olcSpSessionlog', 'Number of write operations keep in session log',
        value_format='%s operation(s)', default='disabled',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcSpNoPresent', 'Skip Present phase of refreshing',
        default='FALSE', dash=True, lines=result_lines)
    if 'TRUE' in get_entry_attr(options, overlay, 'olcSpSessionlog'):
        result_lines.append(
            '  => Caution: This value should only be set TRUE for a syncprov '
            'instance on top of a log database  (such as one managed by the '
            'accesslog overlay).')
    audit_attr(
        options, overlay, 'olcSpReloadHint', 'Honor the reloadHint flag in the Sync Control',
        default='FALSE', dash=True, lines=result_lines)
    if (
        'TRUE' not in get_entry_attr(options, overlay, 'olcSpReloadHint')
        and 'accesslog' in overlay_dn
    ):
        result_lines.append(
            '  => Caution: It must be set TRUE when using the accesslog '
            'overlay for delta-based syncrepl replication support.')
    return extended_return(result_lines, lines, line_prefix)


# pylint: disable=unused-argument
def audit_overlay_refint(options, overlay_dn, overlay, lines=None, line_prefix=None):
    """
    Audit overlay refint

    :param options: The script options as argparse.Namespace
    :param overlay_dn: The overlay DN
    :param overlay: The overlay entry
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []
    audit_attr(
        options, overlay, 'olcRefintAttribute', 'Attributes', multiple=True, split=True,
        not_set='No attribute declared',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcRefintNothing',
        '"Nothing" value (=placeholder when the last value of a required attribute is removed)',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcRefintModifiersName', 'Modifier DN',
        default='cn=Referential Integrity Overlay',
        dash=True, lines=result_lines)
    return extended_return(result_lines, lines, line_prefix)


# pylint: disable=unused-argument
def audit_overlay_accesslog(options, overlay_dn, overlay, lines=None, line_prefix=None):
    """
    Audit overlay accesslog

    :param options: The script options as argparse.Namespace
    :param overlay_dn: The overlay DN
    :param overlay: The overlay entry
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []
    audit_attr(
        options, overlay, 'olcAccessLogDB', 'Database used to store logs',
        not_set='not set (/!\\ BUT REQUIRED /!\\)',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcAccessLogOps', 'Logged operation(s)', multiple=True, split=True,
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcAccessLogBase', 'Base logged operation(s)', multiple=True,
        line_return=True, dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcAccessLogOld', 'Logged old-content filter',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcAccessLogOldAttr', 'Logged old-content attributes',
        multiple=True, split=True,
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcAccessLogPurge', 'Purge rule',
        dash=True, lines=result_lines)
    audit_attr(
        options, overlay, 'olcAccessLogSuccess', 'Log only successful requests', default='FALSE',
        dash=True, lines=result_lines)
    return extended_return(result_lines, lines, line_prefix)


def audit_schema(options, entry_dn, entry, line_prefix=None, lines=None):
    """
    Audit a schema entry

    :param options: The script options as argparse.Namespace
    :param entry_dn: The schema DN
    :param entry: The schema entry
    :param line_prefix: If set, all resulting lines will be prefix with this string
    :param lines: If set, it must be a list and the resulting lines will be use to extend this
                  list
    """
    result_lines = []

    if entry_dn == 'cn=schema,cn=config':
        result_lines.append(f'Global schema ({entry_dn}):')
    else:
        schema_rdn = get_entry_attr(options, entry, 'cn', all_values=False)
        schema_idx, schema_name = parse_olc_numbered_value(schema_rdn)
        result_lines.append(f'Schema #{schema_idx} - {schema_name}:')

    audit_attr(
        options, entry, 'olcObjectIdentifier', 'Object identifiers', only_if_set=True,
        multiple=True, line_return=True, dash=True, lines=result_lines)

    def format_syntax(value):
        return re.sub(
            r'\s+(DESC|X-BINARY-TRANSFER-REQUIRED|X-NOT-HUMAN-READABLE)\s+',
            r'\n     \1 ',
            value.replace('\n', ' ')
        )
    audit_attr(
        options, entry, 'olcLdapSyntaxes', 'Syntaxes', only_if_set=True, multiple=True,
        line_return=True, format_value=format_syntax, dash=True, lines=result_lines)

    audit_attr(
        options, entry, 'olcDitContentRules', 'DIT content rules', only_if_set=True,
        multiple=True, line_return=True, dash=True, lines=result_lines)

    def format_attr(attr):
        attr = re.sub(r'\s*\n\s*', ' ', attr)
        # Add line return after each clauses
        attr = re.sub(
            r'\s+(NAME|DESC|SUP|EQUALITY|SUBSTR|ORDERING|SYNTAX|SINGLE-VALUE)\s+',
            r'\n     \1 ', attr)
        return attr
    audit_attr(
        options, entry, 'olcAttributeTypes', 'Attributes', not_set='no defined attribute',
        multiple=True, line_return=True, format_value=format_attr, dash=True, lines=result_lines)

    def format_oc(objectclass):
        # Remove existing line returns
        objectclass = re.sub(r'\s*\n\s*', ' ', objectclass)
        # Add line return after each clauses
        objectclass = re.sub(
            r'\s+(NAME|DESC|SUP|ABSTRACT|STRUCTURAL|AUXILIARY)\s+',
            r'\n     \1 ', objectclass)
        objectclass = re.sub(
            r'\s+(MUST|MAY)\s+',
            r'\n     \1 ', objectclass)
        return objectclass
    audit_attr(
        options, entry, 'olcObjectClasses', 'ObjectClasses', not_set='no defined ObjectClass',
        multiple=True, line_return=True, format_value=format_oc, dash=True, lines=result_lines)

    return extended_return(result_lines, lines, line_prefix)


def main(argv=None):
    """ Entry point of the script """
    parser = get_opts_parser(desc=__doc__, progress=True)

    add_source_options(parser)

    parser.add_argument(
        '-b', '--base',
        action="store",
        dest="base",
        help=f'Base DN of the OLC database (default: {DEFAULT_BASEDN})',
        type=str,
        default=DEFAULT_BASEDN
    )
    argcomplete.autocomplete(parser)
    options = parser.parse_args(argv)
    check_source_options(parser, options)

    init_logging(options, "Audit OpenLDAP configuration")

    entries = get_entries(options)

    root_entry = entries.get('cn=config')
    if not root_entry:
        parser.error(
            'cn=config entry not found. Seem not to be an OpenLDAP configuration directory.')

    lines = []
    audit_attr(
        options, root_entry, 'olcConfigDir', 'Configuration directory', only_if_set=True,
        lines=lines)
    audit_attr(
        options, root_entry, 'olcConfigFile', 'Configuration file', only_if_set=True,
        lines=lines)
    audit_attr(options, root_entry, 'olcLogLevel', 'Log levels', default='disabled', lines=lines)
    audit_attr(options, root_entry, 'olcLogFile', 'Log file', only_if_set=True, lines=lines)
    audit_attr(options, root_entry, 'olcPluginLogFile', 'Slapi plugins log file',
               only_if_set=True, lines=lines)
    audit_attr(options, root_entry, 'olcReverseLookup', 'Enable client name lookup',
               only_if_set=True, lines=lines)
    audit_attr(options, root_entry, 'olcPidFile', 'PID file', only_if_set=True, lines=lines)
    audit_attr(options, root_entry, 'olcServerID', 'Server ID', lines=lines)
    audit_attr(options, root_entry, 'olcReferral', 'Referral URL', only_if_set=True, lines=lines)
    audit_attr(options, root_entry, 'olcReadOnly', 'Read-only mode', only_if_set=True, lines=lines)
    audit_attr(
        options, root_entry, 'olcAllows', 'Allowed deprecated feature(s)', multiple=True,
        not_set='no deprecated feature enabled', only_if_set=True, lines=lines)
    audit_attr(
        options, root_entry, 'olcDisallows', 'Disallowed feature(s)', multiple=True,
        not_set='no disallowed feature', only_if_set=True, lines=lines)
    audit_attr(
        options, root_entry, 'olcRequires', 'Globally required conditions', multiple=True,
        only_if_set=True, lines=lines)
    audit_attr(
        options, root_entry, 'olcRestrict', 'Globally restricted operations', multiple=True,
        only_if_set=True, lines=lines)
    audit_attr(
        options, root_entry, 'olcAttributeOptions', 'Attribute options', multiple=True,
        only_if_set=True, line_return=True, lines=lines)
    audit_attr(
        options, root_entry, 'olcRootDSE',
        'Additional LDIF file containing user defined attributes for the root DSE',
        only_if_set=True, lines=lines)
    audit_attr(
        options, root_entry, 'olcDitContentRules', 'DIT content rule(s)',
        only_if_set=True, line_return=True, lines=lines)
    audit_attr(
        options, root_entry, 'olcGentleHUP', 'Enable "Gentle" shutdown on SIGHUP signal',
        default='FALSE', lines=lines)

    audit_attrs(
        'Timeout',
        audit_attr(
            options, root_entry, 'olcIdleTimeout', 'On Idle client connections',
            default='0 == no timeout'),
        audit_attr(
            options, root_entry, 'olcWriteTimeout', 'On Write operations',
            default='0 == no timeout'),
        lines=lines
    )

    audit_attrs(
        'TLS',
        audit_attr(
            options, root_entry, 'olcTLSCertificateFile', 'Certificate file',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSCertificateKeyFile', 'Certificate key file',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSCACertificateFile', 'CA certificate file',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSCACertificatePath', 'CA certificate path',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSCRLCheck', 'CRL check level',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSCRLCheck', 'CRL file',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSCipherSuite', 'Ciphers suite',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSDHParamFile', 'DH parameters file',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSECName',
            'Elliptic curve (for Diffie-Hellman ephemeral key exchange)',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSProtocolMin', 'Minimum supported protocol',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSRandFile', 'Random file',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTLSVerifyClient', 'Verify client',
            only_if_set=True),

        lines=lines
    )

    audit_attrs(
        'Authentication',
        audit_attr(
            options, root_entry, 'olcAuthIDRewrite', 'Simple login rewrite rule',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcAuthzPolicy', 'Proxy authentication policy',
            only_if_set=True),
        audit_attr(
            options, root_entry, 'olcAuthzRegexp', 'Proxy authentication login regex rewrite rule',
            multiple=True, line_return=True, only_if_set=True),
        lines=lines
    )

    audit_attrs(
        'Security',
        audit_attr(
            options, root_entry, 'olcSecurity', 'Security  Strength Factor (SSF) required',
            multiple=True, split=True, only_if_set=True),
        audit_attr(
            options, root_entry, 'olcLocalSSF',
            'Security  Strength Factor (SSF) for local sessions', only_if_set=True),
        audit_attr(
            options, root_entry, 'olcPasswordHash',
            'Hash format used by OpenLDAP when it hashing passwords itself', only_if_set=True),
        audit_attr(
            options, root_entry, 'olcPasswordCryptSaltFormat',
            'Salt format passed to crypt when generating {CRYPT} passwords', only_if_set=True),
        lines=lines
    )

    audit_attrs(
        'SASL',
        audit_attr(
            options, root_entry, 'olcSaslHost', 'SASL host fqdn', only_if_set=True),
        audit_attr(
            options, root_entry, 'olcSaslRealm', 'SASL realm', only_if_set=True),
        audit_attr(
            options, root_entry, 'olcSaslAuxprops',
            'Auxprop plugins to use for authentication lookups', only_if_set=True),
        audit_attr(
            options, root_entry, 'olcSaslSecProps',
            'Cyrus SASL security properties', only_if_set=True),
        lines=lines
    )

    audit_attrs(
        'Networking',
        audit_attr(
            options, root_entry, 'olcTCPBuffer', 'Custom TCP buffer size', only_if_set=True),
        audit_attr(
            options, root_entry, 'olcSockbufMaxIncoming',
            'Maximum incoming LDAP PDU size for anonymous sessions', only_if_set=True),
        audit_attr(
            options, root_entry, 'olcSockbufMaxIncomingAuth',
            'Maximum incoming LDAP PDU size for authenticated sessions', only_if_set=True),
        audit_attr(
            options, root_entry, 'olcSizeLimit',
            'Global maximum number of entries to return from a search operation', only_if_set=True),
        audit_attr(
            options, root_entry, 'olcTimeLimit',
            'Global maximum number of seconds slapd will spend answering a search request', multiple=True, line_return=True, only_if_set=True),
        audit_attr(
            options, root_entry, 'olcConnMaxPending', 'Maximum pending connections', default=100),
        audit_attr(
            options, root_entry, 'olcConnMaxPendingAuth', 'Maximum pending auth connections',
            default=1000),
        lines=lines
    )

    audit_attrs(
        'Threading',
        audit_attr(
            options, root_entry, 'olcThreads', 'Maximum size of the primary thread pool',
            default=16),
        audit_attr(
            options, root_entry, 'olcToolThreads',
            'Maximum number of threads to use in tool mode', default=1),
        audit_attr(
            options, root_entry, 'olcListenerThreads',
            'Number of threads to use for the connection manager', default=1),
        lines=lines
    )

    audit_attrs(
        'Indexing',
        audit_attr(
            options, root_entry, 'olcIndexIntLen', 'Key length for ordered integer indices',
            default='4'),
        audit_attr(
            options, root_entry, 'olcIndexSubstrIfMaxlen',
            'Maximum  length for subinitial and subfinal indices', default='4'),
        audit_attr(
            options, root_entry, 'olcIndexSubstrIfMinlen',
            'Minimum length for subinitial and subfinal indices', default='2'),
        audit_attr(
            options, root_entry, 'olcIndexSubstrAnyLen',
            'Length used for subany indices', default='4'),
        audit_attr(
            options, root_entry, 'olcIndexSubstrAnyStep',
            'Steps  used  in  subany  index lookups', default='2'),
        lines=lines
    )

    # Modules:
    modules = []
    for entry_dn, entry in get_matching_entries(
            options, entries, filters=[AuditFilter('objectclass=olcModuleList')]):
        modules.extend([
            get_olc_unnumbered_value(value)
            for value in get_entry_attr(options, entry, 'olcModuleLoad')
        ])
    if modules:
        lines.append('Loaded modules:')
        lines.extend([f'  - #{idx} - {module}' for idx, module in enumerate(modules)])
    else:
        lines.append('Loaded modules: no loaded module')

    # Schemas:
    for entry_dn, entry in get_matching_entries(
            options, entries, filters=[AuditFilter('objectclass=olcSchemaConfig')]):
        audit_schema(options, entry_dn, entry, lines=lines)

    # Databases:
    audit_databases(options, entries, lines=lines)
    print('\n'.join(lines))
