""" LDIF file helpers """

from getpass import getpass
import logging

from mylib.ldap import LdapServer


log = logging.getLogger(__name__)
DEFAULT_LDAP_FILTER = '(objectClass=*)'
DEFAULT_LDAP_PAGE_SIZE = 500


def add_ldap_options(parser):
    """ Add LDAP server options to an argparse.ArgumentParser instance """

    ldap_opts = parser.add_argument_group("LDAP options")

    ldap_opts.add_argument(
        '-H', '--ldap-uri',
        action="store",
        type=str,
        dest="ldap_uri",
        help="LDAP host URI"
    )

    ldap_opts.add_argument(
        '-D', '--ldap-binddn',
        action="store",
        type=str,
        dest="ldap_binddn",
        help="LDAP bind DN"
    )

    ldap_opts.add_argument(
        '-w', '--ldap-bindpwd',
        action="store",
        type=str,
        dest="ldap_bindpwd",
        help="LDAP bind password"
    )

    ldap_opts.add_argument(
        '-F', '--ldap-filter',
        action="store",
        type=str,
        dest="ldap_filter",
        help=f'LDAP filter of the search (default: {DEFAULT_LDAP_FILTER})'
    )

    ldap_opts.add_argument(
        '-2', '--ldap-v2',
        action="store_true",
        dest="ldap_v2",
        help="Use the version 2 of the LDAP protocol"
    )

    ldap_opts.add_argument(
        '-P', '--ldap-paged-search',
        action="store_true",
        dest="ldap_paged_search",
        help="Use paged search"
    )

    ldap_opts.add_argument(
        '--ldap-page-size',
        action="store", type=int,
        dest="ldap_page_size",
        help=f'Page size (for paged search, default: {DEFAULT_LDAP_PAGE_SIZE})',
        default=DEFAULT_LDAP_PAGE_SIZE
    )

    return ldap_opts


def check_ldap_options(options):
    """ Check validty of LDAP options in an argparse.Namespace instance """
    if not options.ldap_uri:
        return False

    if options.ldap_binddn and not options.ldap_bindpwd:
        options.ldap_bindpwd = getpass(f'Please enter password for {options.ldap_binddn}:')

    return True


def get_ldap_entries(options):
    """
    Get LDAP entries found

    :param options: The options of the current script (argparse.Namespace)

    return dict of LDAP entries or False if no LDAP source configured
    """

    if not options.ldap_uri:
        return False

    ldap_server = LdapServer(
        options.ldap_uri, options.ldap_binddn, options.ldap_bindpwd,
        v2=options.ldap_v2, raiseOnError=True, logger=log)

    scope = options.scope if 'scope' in options and options.scope else 'sub'
    if isinstance(scope, int):
        if scope == 0:
            scope = 'base'
        elif scope == 1:
            scope = 'one'
        else:
            scope = 'sub'

    log.info('Search entries in LDAP directory...')
    if options.ldap_paged_search:
        return ldap_server.paged_search(
            basedn=options.base,
            filterstr=options.ldap_filter,
            attrs=[],
            scope=scope,
            pagesize=options.ldap_page_size)

    return ldap_server.search(
        basedn=options.base,
        filterstr=options.ldap_filter,
        attrs=[],
        scope=scope)
